import os
import json
import time
import numpy as np
from collections import deque
from scipy.spatial.distance import euclidean
from keras.models import Sequential
from keras.layers import LSTM, Dense
from tensorflow.keras.models import load_model
from keras.models import model_from_json
import threading
from queue import Full
import queue

# Model Settings
script_dir = os.path.dirname(os.path.realpath(__file__))
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # Get project root path
mdl_dir = os.path.join(project_root, 'mdl')  # Path to mdl directory

# Loading Model Paths
model_path = os.path.join(mdl_dir, 'model_architecture.json')
model_weights_path = os.path.join(mdl_dir, 'model_weights.keras')

# loading Model Saved in keras
#model_path = os.path.join(script_dir, '..', 'mdl', 'model_arch.keras')
#model_weights_path = os.path.join(script_dir, '..', 'mdl', 'model_weights.keras')

print(f'Model Path: {model_path}')
print(f'Model Weights Path: {model_weights_path}')

# Load the model architecture from JSON
with open(model_path, "r") as json_file:
    loaded_model_json = json_file.read()
model = model_from_json(loaded_model_json)

# Load keras model
#model = load_model(model_path)

# Loading Weights
model.load_weights(model_weights_path)
print(f"Model loaded: {model.summary()}")

def preprocess_data(input_data):
    """
    Preprocess the input data to extract the positions and distances of body parts to the normalization target (Defaults to Right Foot).
    Args:
        input_data (dict): dictionary containing the positions of body parts extracted from a JSON file.
    Returns:
        positions (list): list of positions of body parts.
        distances (list): list of distances of body parts to the normalization target.
    """

    body_parts = [
        'Pelvis', 'L5', 'L3', 'T12', 'T8', 'Neck', 'Head',
        'Right Shoulder', 'Right Upper Arm', 'Right Forearm', 'Right Hand',
        'Left Shoulder', 'Left Upper Arm', 'Left Forearm', 'Left Hand',
        'Right Upper Leg', 'Right Lower Leg', 'Right Toe',
        'Left Upper Leg', 'Left Lower Leg', 'Left Foot', 'Left Toe'
        ]

    right_foot_position = None
    positions = []
    distances = []
    
    right_foot = next((item for item in input_data['list'] if item['name'] == 'Right Foot'), None)
    right_foot_position = np.array([right_foot['position']['x'], right_foot['position']['y'], right_foot['position']['z']]) if right_foot else None

    # normalize data to the right foot
    for part in body_parts:
        for item in input_data['list']:
            if item['name'] == part:
                part_position = np.array([item['position']['x'], item['position']['y'], item['position']['z']])
                positions.extend(part_position)
                if right_foot_position is not None:
                    distance = np.linalg.norm(part_position - right_foot_position)
                    distances.append(distance)
                break
        else:
            # if a body part is not found, add dummy position and distance
            positions.extend([0, 0, 0])
            distances.append(0)

    return positions, distances


def predict_frames(frames_array: np.array, window_size: int = 5, verbose: bool = False):
    """
    Predict the class of a sequence of frames.

    Args:
        frames_array (np.array): numpy array of shape (1, window_size, num_features)
        window_size (int): number of frames in the sequence
    Returns:
        prediction_class (int): predicted class of the sequence of frames
        predicted_class_label (str): label of the predicted class
    """
    
    prediction = model.predict(frames_array)
    class_labels = ['Socar', 'Chicotear', 'Pressionar', 'Pontuar', 'Sacudir', 'Torcer', 'Deslizar', 'Flutuar']
    prediction_class = np.argmax(prediction)
    predicted_class_label = class_labels[prediction_class]

    if verbose:
        print(f"Predictions: {prediction}")
        print(f"Predicted class: {predicted_class_label}")
        #print("Frames array shape:", frames_array.shape)
    
    return prediction_class, predicted_class_label

def process_file(file_name_in: str, file_name_out: str, window_size: int = 5, rate: float = 0.2):
    """
    Process a file containing JSON data and predict the class of a sequence of frames.

    Output the prediction of the sequence of frames, discarding old frames
    if processing falls behind.

    Args:
        file_name_in (str): the name of the file containing JSON data.
        file_name_out (str): the name of the file to write the predictions to.
        window_size (int): number of frames in the sequence.
        rate (float): rate at which the frames are classified.
    """

    # Create a queue to store frames for asynchronous prediction
    prediction_queue = queue.Queue(maxsize=window_size)

    def read_file_and_queue(file):
        """
        Reads frames from the file and adds them to the prediction queue.

        Args:
            file (file object): The file object to read from.
        """
        while True:
            line = file.readline()
            if not line:
                time.sleep(0.1)
                continue
            try:
                input_data = json.loads(line)
                positions, distances = preprocess_data(input_data)

                # Discard old frame if queue is full
                try:
                    prediction_queue.put(distances, timeout=0.1)  # Non-blocking put
                except Full:
                    # Handle discarded frame
                    pass

            except Exception as e:
                print("Error processing data:", e)

    def process_and_predict(queue: queue.Queue):
        """
        Continuously retrieve frames from the prediction queue, create the
        frames array, and perform prediction using a separate thread.

        Args:
            queue (queue.Queue): The prediction queue containing distance data.
        """
        queue_empty_handled = False
        while True:
            time.sleep(rate)
            if queue.qsize() == window_size:
                try:
                    frames_array = np.array([queue.get() for _ in range(queue.qsize())]).reshape(1, window_size, 22)
                    predicted_class, predicted_label = predict_frames(frames_array, window_size, verbose=False)
                    #count_predictions += 1

                    #print("» Frame queue:", count_predictions, end='\n\n')
                    print(f"Prediction: [{predicted_class}] » {predicted_label}", end='\n\n')

                    data = {"predictedClass": int(predicted_class)}
                    with open(file_name_out, 'w') as f:
                        json.dump(data, f)

                    queue_empty_handled = False

                except queue.Empty:
                    # Handle empty queue (optional logging or pausing)
                    print("Queue empty")
                    pass
            else:
                if not queue_empty_handled:
                    predicted_class = -1
                    print(f"Prediction: [{predicted_class}] » No data available", end='\n\n')
                    data = {"predictedClass": int(predicted_class)}
                    with open(file_name_out, 'w') as f:
                        json.dump(data, f)
                    queue_empty_handled = True


    with open(file_name_in, 'r') as file:
        # Start threads for asynchronous reading and prediction
        file_reader_thread = threading.Thread(target=read_file_and_queue, args=(file,))
        prediction_thread = threading.Thread(target=process_and_predict, args=(prediction_queue,))
        file_reader_thread.start()
        prediction_thread.start()

        # Wait for threads to finish (optional, especially if processing continues)
        file_reader_thread.join()
        prediction_thread.join()

if __name__ == "__main__":
    
    # Default values - adjust rate accordingly
    path = script_dir
    # file_name_in = script_dir + "\input_data.json"
    # file_name_out = script_dir + "\predicted_data.json"
    file_name_in = "C:/Users/siopa/OneDrive - Universidade de Lisboa/Ghostdance/Ghostdance/Assets/Json/body_positions.json"
    file_name_out = "C:/Users/siopa/OneDrive - Universidade de Lisboa/Ghostdance/Ghostdance/Assets/Json/predicted_data.json"
    window_size = 5
    rate = 0.4
    process_file(file_name_in, file_name_out, window_size, rate)