import os
import json
import time
import numpy as np
from collections import deque
from scipy.spatial.distance import euclidean
from keras.models import Sequential
from keras.layers import LSTM, Dense
from tensorflow.keras.models import load_model
from keras.models import model_from_json

# Model Settings
script_dir = os.path.dirname(os.path.realpath(__file__))
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # Get project root path
mdl_dir = os.path.join(project_root, 'mdl')  # Path to mdl directory

# Loading Model Paths
model_path = os.path.join(mdl_dir, 'model_architecture.json')
model_weights_path = os.path.join(mdl_dir, 'model_weights.keras')

# loading Model Saved in keras
#model_path = os.path.join(script_dir, '..', 'mdl', 'model_arch.keras')
#model_weights_path = os.path.join(script_dir, '..', 'mdl', 'model_weights.keras')

print(f'Model Path: {model_path}')
print(f'Model Weights Path: {model_weights_path}')

# Load the model architecture from JSON
with open(model_path, "r") as json_file:
    loaded_model_json = json_file.read()
model = model_from_json(loaded_model_json)

# Load keras model
#model = load_model(model_path)

# Loading Weights
model.load_weights(model_weights_path)
print(f"Model loaded: {model.summary()}")

def preprocess_data(input_data):
    """
    Preprocess the input data to extract the positions and distances of body parts to the normalization target (Defaults to Right Foot).
    Args:
        input_data (dict): dictionary containing the positions of body parts extracted from a JSON file.
    Returns:
        positions (list): list of positions of body parts.
        distances (list): list of distances of body parts to the normalization target.
    """

    body_parts = [
        'Pelvis', 'L5', 'L3', 'T12', 'T8', 'Neck', 'Head',
        'Right Shoulder', 'Right Upper Arm', 'Right Forearm', 'Right Hand',
        'Left Shoulder', 'Left Upper Arm', 'Left Forearm', 'Left Hand',
        'Right Upper Leg', 'Right Lower Leg', 'Right Toe',
        'Left Upper Leg', 'Left Lower Leg', 'Left Foot', 'Left Toe'
        ]

    right_foot_position = None
    positions = []
    distances = []
    
    right_foot = next((item for item in input_data['list'] if item['name'] == 'Right Foot'), None)
    right_foot_position = np.array([right_foot['position']['x'], right_foot['position']['y'], right_foot['position']['z']]) if right_foot else None

    # normalize data to the right foot
    for part in body_parts:
        for item in input_data['list']:
            if item['name'] == part:
                part_position = np.array([item['position']['x'], item['position']['y'], item['position']['z']])
                positions.extend(part_position)
                if right_foot_position is not None:
                    distance = np.linalg.norm(part_position - right_foot_position)
                    distances.append(distance)
                break
        else:
            # if a body part is not found, add dummy position and distance
            positions.extend([0, 0, 0])
            distances.append(0)

    return positions, distances


def predict_frames(frames_array: np.array, window_size: int = 5, verbose: bool = False):
    """
    Predict the class of a sequence of frames.

    Args:
        frames_array (np.array): numpy array of shape (1, window_size, num_features)
        window_size (int): number of frames in the sequence
    Returns:
        prediction_class (int): predicted class of the sequence of frames
        predicted_class_label (str): label of the predicted class
    """
    
    prediction = model.predict(frames_array)
    class_labels = ['Socar', 'Chicotear', 'Pressionar', 'Pontuar', 'Sacudir', 'Torcer', 'Deslizar', 'Flutuar']
    prediction_class = np.argmax(prediction)
    predicted_class_label = class_labels[prediction_class]

    if verbose:
        print(f"Predictions: {prediction}")
        print(f"Predicted class: {predicted_class_label}")
        #print("Frames array shape:", frames_array.shape)
    
    return prediction_class, predicted_class_label

def process_file(filename: str, window_size: int = 5, rate: float = 0.1):
    """
    Process a file containing JSON data and predict the class of a sequence of frames.

    Output the prediction of the sequence of frames.

    Args:
        filename (str): name of the file containing the JSON data.
        window_size (int): number of frames in the sequence.
        rate (float): rate at which the file is read.
    """

    # deque to accumulate frames
    frame_queue = deque(maxlen=window_size)
    count_predictions = 0
    
    with open(filename, 'r') as file:
        # continuously read from the file
        while True:
            
            time.sleep(rate)
            line = file.readline()
            
            # if no new data is available, wait and continue
            if not line:
                time.sleep(0.1)
                continue
            try:
                # parse the JSON data
                input_data = json.loads(line)
                
                # preprocess the input data
                positions, distances = preprocess_data(input_data)
                
                frame_queue.append(distances)
                if len(frame_queue) == window_size:
                    
                    # predict using accumulated frames
                    frames_array = np.array(frame_queue).reshape(1, window_size, len(frame_queue[0]))
                    predicted_class, predicted_label = predict_frames(frames_array, window_size, verbose=False)
                    
                    print("» Frame queue:", count_predictions, end='\n\n')
                    print(f"Prediction: [{predicted_class}] » {predicted_label}", end='\n\n')
                    
                    # drop the oldest frame from the queue
                    count_predictions += 1
                    frame_queue.popleft()

            except Exception as e:
                print("Error processing data:", e)

if __name__ == "__main__":
    
    # Default values - adjust rate accordingly
    filename = os.path.join(script_dir, './input_data.json')
    #filename = "C:/Users/siopa/OneDrive - Universidade de Lisboa/Ghostdance/Ghostdance/Assets/Json/body_positions.json"
    window_size = 5
    rate = 0.2
    process_file(filename, window_size, rate)